#include "ctest1.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

char configFileName[MAX_LEN_CONFIG_FILE];
char fullConfigFileName[MAX_LEN_FULL_CONFIG_FILE];
struct TParamCTest1 params;
struct TMil1553Channel* channels;
int channelsCount;
struct TMil1553Device* devices;
int devicesCount;
struct SConfigCh configChannels;  ///< конфигкрация готовности каналов
struct TStatCTest1 statTest;  ///< статистика теста

static unsigned int isBcStopped;

#define PRINT_LINE() LOGGER_PRINT("%s\n", "--- --- ---")
/// \brief стартовый адрес инструкции КШ
#define BC_START_ADDRESS 0
/// \brief групповая команда синхронизации
#define TR_SYNCRO 9

int main(int argc, char* argv[]) {
  initLogCTest1();
  LOGGER_PRINT("%s %s\n", "ctest1 - version", VERSION_APP);
  int ret = GOOD_CODE;
  if (argc != 2) {  // проверка кол-ва аргументов
  usage:
    LOGGER_PRINT("%s\n", "Использование: ctest1 <configfilename>");
    ret = BAD_CODE;
  } else {  // кол-во аргументов - ок
    if (sscanf(argv[1], "%s", configFileName) !=
        1) {  // чтение имени конфигурационного файла
      LOGGER_PRINT("%s\n", "неверный configfilename\n");
      goto usage;
    } else {  //имя конфигурационного файла - ок
      constructFullConfigFileName(fullConfigFileName, MAX_LEN_FULL_CONFIG_FILE,
                                  configFileName);
      initDefaultParamsCTest1(&params);
      ret = readConfigFile(
          fullConfigFileName,
          &params);  // чтение параметров из конфигурационного файла
      if (ret == BAD_CODE) {  //параметры из конфигурационного файла - error
        LOGGER_PRINT("Невозможно прочитать параметры теста из '%s' \n",
                     fullConfigFileName);
      } else {  //параметры из конфигурационного файла - ок
        procCTest1(params);
      }
    }
  }
  LOGGER_FLUSH();
  LOGGER_CLOSE();
  return ret;
}
/// \brief размер мем файла
#define MEM_SIZE 1024
/// \brief инструкции КШ
static unsigned int memDesc[MEM_SIZE];
/// \brief операции КШ
static unsigned int memOp[MEM_SIZE];
/// \brief данные КШ
static unsigned int memData[MEM_SIZE];

void loadMemFiles(void) {
  loadMemTxtFile("./mem/consol_dat_ram.mem", memData, MEM_SIZE);
  loadMemTxtFile("./mem/consol_inst_ram.mem", memDesc, MEM_SIZE);
  loadMemTxtFile("./mem/consol_op_ram.mem", memOp, MEM_SIZE);
}
/// \brief кол-во подциклов в мем файле инструкций
#define MAX_DESC_SHIFT_COUNTER 6
/// \brief стартовые адреса подциклов
unsigned int cycleAddresses[MAX_DESC_SHIFT_COUNTER] = {0x10, 0x20, 0x30,
                                                       0x40, 0x50, 0x60};

void memDescRtOff(unsigned int* word) {
  int countOfOnes = 0;
  *word = (1 << 20) | *word;
  for (int i = 16; i <= 30; i++) {
    if (*word & (1 << i)) {
      countOfOnes++;
    }
  }
  if ((countOfOnes % 2) == 0)
    *word = *word | (1u << 31);
  else
    *word = *word & ~(1u << 31);
}

void memDescRtOffWithCheckReady(int num_rt, int counter) {
  if ((FALSE_VAL == configChannels.mIsReady[num_rt]) &&
      (FALSE_VAL == configChannels.mIsRemote[num_rt])) {
    unsigned int adr = cycleAddresses[counter] + num_rt - 1;
    memDescRtOff(&(memDesc[adr]));
  }
}

void applyParametresToMem(void) {
  int counter;
  if (0 == params.mLine) {  // off line b
    memDescRtOff(&memDesc[0x6]);
    memDescRtOff(&memDesc[0x7]);
    memDescRtOff(&memDesc[0x8]);
    memDescRtOff(&memDesc[0x9]);
  } else {  // off line a
    memDescRtOff(&memDesc[0x2]);
    memDescRtOff(&memDesc[0x3]);
    memDescRtOff(&memDesc[0x4]);
    memDescRtOff(&memDesc[0x5]);
  }
  for (counter = 0; counter < MAX_DESC_SHIFT_COUNTER; counter++) {
    // применение параметров к мем descriptors
    memDescRtOffWithCheckReady(1, counter);
    memDescRtOffWithCheckReady(2, counter);
    memDescRtOffWithCheckReady(3, counter);
    memDescRtOffWithCheckReady(4, counter);
    memDescRtOffWithCheckReady(5, counter);
    memDescRtOffWithCheckReady(6, counter);
    memDescRtOffWithCheckReady(7, counter);
    memDescRtOffWithCheckReady(8, counter);
  }
}

void writeMemToBc(void) {
  if (configChannels.mIsReady[0]) {
    int isGood;
    int fd = mil1553_open(params.mNames[0]);
    isGood = mil1553_bc_writeDataToRam(fd, Mil1553BcTypeInstructions, memDesc,
                                       MEM_SIZE);
    LOGGER_PRINT_STR("./mem/consol_inst_ram.mem");
    if (isGood == GOOD_CODE) {
      LOGGER_PRINT_STR(" записан.\n");
    } else {
      LOGGER_PRINT_STR(" НЕ записан.\n");
    }
    isGood =
        mil1553_bc_writeDataToRam(fd, Mil1553BcTypeOperations, memOp, MEM_SIZE);
    LOGGER_PRINT_STR("./mem/consol_op_ram.mem");
    if (isGood == GOOD_CODE) {
      LOGGER_PRINT_STR(" записан.\n");
    } else {
      LOGGER_PRINT_STR(" НЕ записан.\n");
    }
    isGood =
        mil1553_bc_writeDataToRam(fd, Mil1553BcTypeData, memData, MEM_SIZE);
    LOGGER_PRINT_STR("./mem/consol_dat_ram.mem");
    if (isGood == GOOD_CODE) {
      LOGGER_PRINT_STR(" записан.\n");
    } else {
      LOGGER_PRINT_STR(" НЕ записан.\n");
    }
    mil1553_close(fd);
  }
}

void initChannel_switchOffChannel(int fd) {
  unsigned int opCh = MIL_DEV_CHANNEL_OFF;
  mil1553_manChannel(fd, &opCh);
  unsigned int opDma = MIL1553_OFF;
  mil1553_manDma(fd, opDma);
  mil1553_clearDma(fd);
}

void initChannel_config(int fd, int index) {
  unsigned int modeBus;                            // режим
  unsigned int bus = MIL_BUS_A_EN | MIL_BUS_B_EN;  // шина
  if (0 == index) {
    modeBus = MIL_MODE_BUS_CONTR;  // режим КШ
  } else {
    modeBus = MIL_MODE_TERMINAL_DEV;
  }
  mil1553_setRole(fd, modeBus);
  mil1553_switchBus(fd, bus);
  if (0 != index) {
    mil1553_rt_setAddress(fd, index);
  }
}

void initChannel_initRtSubaddresses(int fd) {
  mil1553_rt_manSubaddresses(
      fd, MIL_RT_BUF_RECEIVE, 0x7FFFFFFE /* меняем с 1 по 30 подадрес*/,
      0x7FFFFFFE /* включаем на приём с 1 по 30 подадрес*/);
  mil1553_rt_manSubaddresses(
      fd, MIL_RT_BUF_TRANSMIT, 0x7FFFFFFE /* меняем с 1 по 30 подадрес*/,
      0x7FFFFFFE /* включаем на приём с 1 по 30 подадрес*/);
}

void initChannel_initRtSubaddressBuffer(int fd) {
  unsigned char subaddress;
  for (subaddress = 1; subaddress <= 30; subaddress++) {
    mil1553_rt_manSubadressTransmitBufferMode(fd, subaddress, MIL_RT_MODE_PROG);
    mil1553_rt_manSubaddressTransmitBufferReady(fd, subaddress,
                                                MIL_RT_BUF_0_ON_1_OFF);
  }
  generateDataRt(fd);
}

void initChannel_switchOnChannel(int fd) {
  // включаем ДМА
  unsigned int opDma = MIL1553_ON;
  mil1553_manDma(fd, opDma);
  // включаем канал
  unsigned int opCh = MIL_DEV_CHANNEL_ON;
  mil1553_manChannel(fd, opCh);
}

void initChannels(void) {
  applyParametresToMem();  // применение парметров к мем файлу с инструкциями
  int index;
  for (index = 0; index < COUNT_CH; index++) {  // цикл по каналам
    if (configChannels.mIsReady[index]) {
      int fd = mil1553_open(params.mNames[index]);
      initChannel_switchOffChannel(fd);  // выкл канала
      initChannel_config(fd, index);  // конфигурация канала
      if (0 != index) {               // для ОУ
        initChannel_initRtSubaddresses(fd);  // установка разрешений подадресов
        initChannel_initRtSubaddressBuffer(
            fd);  // установка буферов на передачу
      }
      initChannel_switchOnChannel(fd);  // вкл канала
      mil1553_close(fd);
    }
  }
  writeMemToBc();  // запись мем файлов в ОЗУ КШ
}

void deinitChannels(void) {
  int index;
  for (index = 0; index < COUNT_CH; index++) {  // цикл по каналам
    if (configChannels.mIsReady[index]) {
      int fd = mil1553_open(params.mNames[index]);
      // деинициализация
      unsigned int opCh = MIL_DEV_CHANNEL_OFF;
      mil1553_manChannel(fd, &opCh);
      unsigned int opDma = MIL1553_OFF;
      mil1553_manDma(fd, opDma);
      mil1553_clearDma(fd);
      mil1553_close(fd);
    }
  }
}

void startBc(void) {
  if (configChannels.mIsReady[0]) {  // старт канала КШ, если он есть
    int fd = mil1553_open(params.mNames[0]);
    int numCh = mil1553ud_getNumberChannel(fd);
    int shift = 0;
    switch (numCh) {
      case CH_NUM_1:
        shift = 10;
        break;
      case CH_NUM_2:
        shift = 14;
        break;
      case CH_NUM_3:
        shift = 18;
        break;
      case CH_NUM_4:
        shift = 22;
        break;
    }
    SADDR_DATA reg;
    reg.daddr = INTERRUPT_MASK;
    reg.data = 1 << shift;
    if (params.mBcInt) mil1553_writeReg(fd, &reg);
    mil1553ud_bcSetStartInstructionAddress(fd, BC_START_ADDRESS);
    mil1553ud_bcProgram(
        fd, BC_PROGRAM_START);  // BCSTRT - старт выполнения программы КШ;
    mil1553_close(fd);
  }
}

void stopBc(void) {
  if (configChannels.mIsReady[0]) {  // старт канала КШ, если он есть
    int fd = mil1553_open(params.mNames[0]);
    mil1553ud_bcProgram(
        fd, BC_PROGRAM_STOP);  // BCSTRT - старт выполнения программы КШ;
    mil1553_close(fd);
  }
}

void constructFullConfigFileName(char* fullname, int size, char* name) {
  snprintf(fullname, size, "%s%s", STR_CONFIG, name);
}

#define READ_CHANNEL_NAME(num)                                          \
  sscanf(str, "%s %s", typeArgStr, valStr);                             \
  configChannels.mIsRemote[num] = FALSE_VAL;                            \
  if (0 == strncmp(valStr, NULL_DEV, sizeof(NULL_DEV))) {               \
    configChannels.mIsReady[num] = FALSE_VAL;                           \
    memcpy(params->mNames[num], valStr, CHANNEL_NAME_LENGTH);           \
  } else if (0 == strncmp(valStr, REMOTE_DEV, sizeof(REMOTE_DEV))) {    \
    configChannels.mIsReady[num] = FALSE_VAL;                           \
    configChannels.mIsRemote[num] = TRUE_VAL;                           \
    memcpy(params->mNames[num], valStr, CHANNEL_NAME_LENGTH);           \
  } else {                                                              \
    char fullNameCh[CHANNEL_NAME_LENGTH];                               \
    configChannels.mIsReady[num] = TRUE_VAL;                            \
    snprintf(fullNameCh, CHANNEL_NAME_LENGTH, "%s%s", STR_DEV, valStr); \
    memcpy(params->mNames[num], fullNameCh, CHANNEL_NAME_LENGTH);       \
  }

int readConfigFile(char* fullname, struct TParamCTest1* params) {
  FILE* file = fopen(fullname, "r");
  char* str;
  if (file == NULL) {
    LOGGER_PRINT("Невозможно открыть конфигурационный файл '%s'\n", fullname);
    return BAD_CODE;
  }
  LOGGER_PRINT("парсинг '%s':\n", fullname);
  while (!feof(file)) {
    char valStr[CHANNEL_NAME_LENGTH];
    readline(file, &str);
    LOGGER_PRINT("%s\n", str);  // for debug
    char typeArgStr[30];
    int valArg = BAD_CODE;
    sscanf(str, "%s %s", typeArgStr, valStr);
    if (0 ==
        strncmp(typeArgStr, T_CTEST1_DURATION_H, sizeof(T_CTEST1_DURATION_H))) {
      sscanf(str, "%s %d", typeArgStr, &valArg);
      params->mDurationHours = valArg;
    } else if (0 == strncmp(typeArgStr, T_CTEST1_DURATION_M,
                            sizeof(T_CTEST1_DURATION_M))) {
      sscanf(str, "%s %d", typeArgStr, &valArg);
      params->mDurationMinutes = valArg;
    } else if (0 == strncmp(typeArgStr, T_CTEST1_LINE, sizeof(T_CTEST1_LINE))) {
      sscanf(str, "%s %d", typeArgStr, &valArg);
      params->mLine = valArg;
    } else if (0 ==
               strncmp(typeArgStr, T_CTEST1_BCINT, sizeof(T_CTEST1_BCINT))) {
      sscanf(str, "%s %d", typeArgStr, &valArg);
      params->mBcInt = valArg;
    } else if (0 == strncmp(typeArgStr, T_CTEST1_BC, sizeof(T_CTEST1_BC))) {
      READ_CHANNEL_NAME(0);
    } else if (0 == strncmp(typeArgStr, T_CTEST1_RT1, sizeof(T_CTEST1_RT1))) {
      READ_CHANNEL_NAME(1);
    } else if (0 == strncmp(typeArgStr, T_CTEST1_RT2, sizeof(T_CTEST1_RT2))) {
      READ_CHANNEL_NAME(2);
    } else if (0 == strncmp(typeArgStr, T_CTEST1_RT3, sizeof(T_CTEST1_RT3))) {
      READ_CHANNEL_NAME(3);
    } else if (0 == strncmp(typeArgStr, T_CTEST1_RT4, sizeof(T_CTEST1_RT4))) {
      READ_CHANNEL_NAME(4);
    } else if (0 == strncmp(typeArgStr, T_CTEST1_RT5, sizeof(T_CTEST1_RT5))) {
      READ_CHANNEL_NAME(5);
    } else if (0 == strncmp(typeArgStr, T_CTEST1_RT6, sizeof(T_CTEST1_RT6))) {
      READ_CHANNEL_NAME(6);
    } else if (0 == strncmp(typeArgStr, T_CTEST1_RT7, sizeof(T_CTEST1_RT7))) {
      READ_CHANNEL_NAME(7);
    } else if (0 == strncmp(typeArgStr, T_CTEST1_RT8, sizeof(T_CTEST1_RT8))) {
      READ_CHANNEL_NAME(8);
    }
    memset(typeArgStr, 0, 30);
  }
  fclose(file);
  LOGGER_PRINT(
      "ПАРАМЕТРЫ: duration_h= %d, duration_m= %d, line= %d, bc= %s,\n rt1= "
      "%s,\n rt2= %s,\n rt3= %s,\n rt4= %s,\n rt5= %s,\n rt6= %s,\n rt7= %s,\n "
      "rt8= %s,\n bcint= %d \n\n",
      params->mDurationHours, params->mDurationMinutes, params->mLine,
      params->mNames[0], params->mNames[1], params->mNames[2],
      params->mNames[3], params->mNames[4], params->mNames[5],
      params->mNames[6], params->mNames[7], params->mNames[8], params->mBcInt);
  PRINT_LINE();
  return GOOD_CODE;
}

void initLogCTest1(void) {
  char filename[LOGGER_FILENAME_LENGTH];
  constructLogFilename(filename, LOGGER_FILENAME_LENGTH, "ctest1");
  LOGGER_OPEN(filename);
}

void initDefaultParamsCTest1(struct TParamCTest1* params) {
  params->mDurationHours = 0;  ///< продолжительность теста - часы
  params->mDurationMinutes = 0;  ///< продолжительность теста - минуты
  params->mBcInt = 0;
}

void initTestDuration(struct TParamCTest1* params,
                      unsigned int* allDurationSeconds,
                      unsigned int* curDurationSeconds) {
  *curDurationSeconds = 0;
  *allDurationSeconds =
      params->mDurationMinutes * 60 + params->mDurationHours * 60 * 60;
}

static unsigned int dataAdrRtLineA[8] = {0x0,  0x10, 0x20, 0x30,
                                         0x40, 0x50, 0x60, 0x70};
static unsigned int dataAdrRtLineB[8] = {0x80, 0x90, 0xA0, 0xB0,
                                         0xC0, 0xD0, 0xE0, 0xF0};

unsigned int isGoodDataRt(struct TMIL1553_DMA_BLOCK* block, int index_ch) {
  if (BC_CH_INDEX != index_ch) {  // BC
                                  // проверка данных ОУ
    unsigned short* dataWord16 =
        (unsigned short*)mil1553ud_dmaBlock_getPtrDataPart(block);
    unsigned short* dataWord16Etalon =
        (params.mLine == 0) ? &(memData[dataAdrRtLineA[index_ch - 1]])
                            : &(memData[dataAdrRtLineB[index_ch - 1]]);
    for (unsigned int i = 0; i < 15; i++) {
      if (dataWord16[i] != dataWord16Etalon[i]) return MIL1553UD_FALSE;
    }
  }
  return MIL1553UD_TRUE;
}

unsigned int isGoodData(struct TMIL1553_DMA_BLOCK* block, int index_ch) {
  if (BC_CH_INDEX == index_ch) {  // BC
    // проверка данных КШ
    unsigned short* dataWord16 =
        (unsigned short*)mil1553ud_dmaBlock_getPtrDataPart(block);
    for (unsigned int i = 0; i < 7; i++) {
      if (dataWord16[i] != CHESS_DATA) return MIL1553UD_FALSE;
    }
  }
  return MIL1553UD_TRUE;
}

int isRun(int fd) {
  SADDR_DATA regBcCondCode;
  unsigned int numChannel = mil1553ud_getNumberChannel(fd);
  regBcCondCode.daddr = MIL_REG(numChannel, GPF_BC_COND_CODE_CH);
  mil1553_readReg(fd, &regBcCondCode);
  return (regBcCondCode.data & (1u << 31)) ? MIL1553UD_TRUE : MIL1553UD_FALSE;
}

char* stringsTransaction[11] = {"формат 0",
                                "формат 1 КШ -> ОУ (КС)",
                                "формат 2 ОУ -> КШ (КС)",
                                "формат 3 ОУ -> ОУ (КС)",
                                "формат 4 КУ",
                                "формат 5 КУ с СД (ОУ -> КШ)",
                                "формат 6 КУ с СД (КШ -> ОУ)",
                                "формат 7 КШ -> ОУ групповая  (КС)",
                                "формат 8 ОУ -> ОУ групповая   (КС)",
                                "формат 9 КУ групповая",
                                "формат 10 КУ c СД (КШ -> ОУ) групповая"};

char* getMilTransactionString(unsigned int typeTransaction) {
  return stringsTransaction[typeTransaction];
}

char* errorsBcString[16] = {"Ошибка таймаута приёма.",
                            "Ошибка декодера Манчестера II.",
                            "Ошибка синхронизации.",
                            "Ошибка последовательности.",
                            "Интервал между сообщениями t1 меньше 4 мкс.",
                            "Ошибка разрыва данных.",
                            "Ошибка чётности адреса ОУ.",
                            "КС2 содержит КУ.",
                            "Ошибка КУ.",
                            "Длина сообщения КС1 /= длине сообщения КС2.",
                            "Адрес ОУ1 = ОУ2 (команды ОУ-ОУ).",
                            "КУ зарезервирована.",
                            "КУ не определена.",
                            "Адрес/подадрес запрещён.",
                            "Установлен запрет на выполнение принятой КУ.",
                            "Принято КС с групповым адресом и битом передача "
                            "установленным в 1 (кроме КУ)."};

void printBcError(char* block) {
  int index;
  unsigned int typeDmaBlock = MIL1553UD_TD_DMA_BLOCK;  // ОУ
  unsigned int countServiceWords16 =
      mil1553ud_dmaBlock_getCountServiceWords16(typeDmaBlock, block);
  unsigned short* block16 = (unsigned short*)block;
  int shift16 = countServiceWords16 - 1;
  unsigned int error = block16[shift16];
  LOGGER_PRINT("%s\n", "Сервисные слова блока ПДП(DMA) в сыром виде:");
  for (index = 0; index < countServiceWords16; index++) {
    LOGGER_PRINT("0x%x, ", block16[index]);
  }
  LOGGER_PRINT("%s", "\n");
  LOGGER_PRINT("Значение регистра ошибок = 0x%x:\n", error);
  for (index = 0; index < 16; index++) {
    if (error & (1 << index)) {
      LOGGER_PRINT("%s;\n", errorsBcString[index]);
    }
  }
}

void processingBc(void) {
  int index = 0;  // инрдекс КШ
  if (configChannels.mIsReady[index]) {
    if (MIL1553UD_FALSE == isBcStopped) {
      int fd = mil1553_open(params.mNames[index]);
      int isRunValue = isRun(fd);
      mil1553_close(fd);
      if (MIL1553UD_FALSE == isRunValue) startBc();
    }
    {
      struct MIL1553_DMA_CONTAINER_BLOCKS container;
      int ret, indexBlk;
      int fd = mil1553_open(params.mNames[index]);
      ret = mil1553_readDma(fd, &container);  // чтение dma блоков
      if (GOOD_CODE == ret) {
        for (indexBlk = 0; indexBlk < container.mCount; indexBlk++) {
          char* block = container.mBlocks[indexBlk].mData;
          unsigned int typeDmaBlock = MIL1553UD_BC_DMA_BLOCK;  // КШ
          unsigned int isSuccessTransaction =
              mil1553ud_dmaBlock_isSuccessTransaction(typeDmaBlock, block);
          unsigned int typeTransaction =
              mil1553ud_dmaBlock_getTypeTransaction(typeDmaBlock, block);
          unsigned int freeTimer =
              mil1553ud_dmaBlock_getFreeTimer(typeDmaBlock, block);
          unsigned int countServiceWords16 =
              mil1553ud_dmaBlock_getCountServiceWords16(typeDmaBlock, block);
          unsigned int countDataWords16 =
              mil1553ud_dmaBlock_getCountDataWords16(typeDmaBlock, block);
          unsigned int activeBus =
              mil1553ud_dmaBlock_getActiveBus(typeDmaBlock, block);
          unsigned short* cw = (unsigned short*)(block + 5 * 4);
          unsigned int isGroupCom = MIL1553UD_FALSE;
          if ((cw[0] >> 11) == 31)  // Групповая команда
            isGroupCom = MIL1553UD_TRUE;
          // накопление статистики
          statTest.mCh[index].mDmaBlk++;
          if (MIL1553UD_FALSE ==
              isSuccessTransaction) {  // проверка на наличие ошибок
            statTest.mCh[index].mErr++;
            PRINT_LINE();
            // печать инфо об ошибке
            LOGGER_PRINT("%s в КШ(BC): ", "Ошибка");
            LOGGER_PRINT(
                "Freetimer= %u, Сообщение= %d (%s), Шина= %s, Сервисные слова= "
                "%d, Слова данных= %d\n",
                freeTimer, typeTransaction,
                getMilTransactionString(typeTransaction),
                (activeBus == MIL1553_BUS_A) ? "А" : "Б", countServiceWords16,
                countDataWords16);
            printBcError(block);
            if ((typeTransaction == 1) | (typeTransaction == 2)) {
              if ((countDataWords16 > 0) && (isSuccessTransaction)) {
                unsigned short* dataWord16 =
                    (unsigned short*)mil1553ud_dmaBlock_getPtrDataPart(block);
                LOGGER_PRINT("%s: ", "Слова данных");
                for (unsigned int i = 0; i < countDataWords16; i++) {
                  if (i == countDataWords16 - 1) {
                    LOGGER_PRINT("0x%x; \n", dataWord16[i]);
                  } else {
                    LOGGER_PRINT("0x%x,", dataWord16[i]);
                  }
                }
              }
            }
          }
          if (TR_SYNCRO == typeTransaction) {
            generateDataRt(fd);
          }
        }
      } else {
        LOGGER_PRINT("Ошибка ПДП(DMA) в КШ(BC) (%s) = %d\n",
                     params.mNames[index], ret);
      }
      mil1553_close(fd);
    }
  }
}

void generateDataRt(int fd) {
  struct TRtSubaddressData data0;
  for (int index = 0; index < 32; index++) {
    data0.mWords[index] = rand();
  }
  data0.mCount = 32;
  unsigned char subaddress;
  for (subaddress = 1; subaddress <= 30; subaddress++) {
    mil1553_rt_writeDataToSubaddressTransmit(fd, MIL_RT_BUF0, subaddress,
                                             data0);
    mil1553_rt_writeDataToSubaddressTransmit(fd, MIL_RT_BUF1, subaddress,
                                             data0);
  }
}

char* errorsRtString[16] = {
    "Ошибка приёма данных.",
    "Ошибка декодера Манчестера II.",
    "Ошибка синхронизации.",
    "Ошибка последовательности.",
    "Интервал между сообщениями t1 меньше 4 мкс.",
    "Ошибка разрыва данных.",
    "Ошибка чётности Манчестер II.",
    "Ошибка таймаута t1.",
    "Выполнено два повтора сообщения.",
    "Выполнен один повтор сообщения.",
    "Нет ответа.",
    "Длина данных, принятых от ОУ не соответствует полю команды.",
    "Ошибка адреса ОУ.",
    "Ошибка формата.",
    "Ошибка длины.",
    "Ошибка самоконтроля данных."};

void printRtError(char* block) {
  int index;
  unsigned int typeDmaBlock = MIL1553UD_TD_DMA_BLOCK;  // ОУ
  unsigned int countServiceWords16 =
      mil1553ud_dmaBlock_getCountServiceWords16(typeDmaBlock, block);
  unsigned short* block16 = (unsigned short*)block;
  int shift16 = countServiceWords16 - 1;
  unsigned int error = block16[shift16];
  LOGGER_PRINT("%s\n", "Сервисные слова блока ПДП(DMA) в сыром виде:");
  for (index = 0; index < countServiceWords16; index++) {
    LOGGER_PRINT("0x%x, ", block16[index]);
  }
  LOGGER_PRINT("%s", "\n");
  LOGGER_PRINT("Значение регистра ошибок = 0x%x:\n", error);
  for (index = 0; index < 16; index++) {
    if (error & (1 << index)) {
      LOGGER_PRINT("%s;\n", errorsRtString[index]);
    }
  }
}

void processingRts(void) {
  int index;
  for (index = 1; index < COUNT_CH; index++) {  // цикл по ОУ
    if (configChannels.mIsReady[index] == TRUE_VAL) {
      struct MIL1553_DMA_CONTAINER_BLOCKS container;
      int ret, indexBlk;
      int fd = mil1553_open(params.mNames[index]);
      ret = mil1553_readDma(fd, &container);  // чтение dma блоков
      if (GOOD_CODE == ret) {
        for (indexBlk = 0; indexBlk < container.mCount; indexBlk++) {
          char* block = container.mBlocks[indexBlk].mData;
          unsigned int typeDmaBlock = MIL1553UD_TD_DMA_BLOCK;  // ОУ
          unsigned int isSuccessTransaction =
              mil1553ud_dmaBlock_isSuccessTransaction(typeDmaBlock, block);
          unsigned int typeTransaction =
              mil1553ud_dmaBlock_getTypeTransaction(typeDmaBlock, block);
          unsigned int freeTimer =
              mil1553ud_dmaBlock_getFreeTimer(typeDmaBlock, block);
          unsigned int countServiceWords16 =
              mil1553ud_dmaBlock_getCountServiceWords16(typeDmaBlock, block);
          unsigned int countDataWords16 =
              mil1553ud_dmaBlock_getCountDataWords16(typeDmaBlock, block);
          unsigned int subaddress = mil1553ud_dmaBlock_tdModeSubaddress(block);
          unsigned int activeBus =
              mil1553ud_dmaBlock_getActiveBus(typeDmaBlock, block);
          // накопление статистики
          statTest.mCh[index].mDmaBlk++;
          if (MIL1553UD_FALSE ==
              isSuccessTransaction) {  // проверка на наличие ошибок
            statTest.mCh[index].mErr++;
            // печать инфо об ошибке
            PRINT_LINE();
            LOGGER_PRINT(
                "Ошибка в ОУ(RT)-%d: Freetimer= %u, Сообщение= %d (%s), Шина= "
                "%s, Подадрес= %d, Сервисные слова= %d, Слова данных= %d\n",
                index, freeTimer, typeTransaction,
                getMilTransactionString(typeTransaction),
                (activeBus == MIL1553_BUS_A) ? "А" : "Б", subaddress,
                countServiceWords16, countDataWords16);
            printRtError(block);
            if (typeTransaction == 1) {
              if ((countDataWords16 > 0) && (isSuccessTransaction)) {
                unsigned short* dataWord16 =
                    (unsigned short*)mil1553ud_dmaBlock_getPtrDataPart(block);
                LOGGER_PRINT("%s: ", "Слова данных");
                for (unsigned int i = 0; i < countDataWords16; i++) {
                  if (i == countDataWords16 - 1) {
                    LOGGER_PRINT("0x%x; \n", dataWord16[i]);
                  } else {
                    LOGGER_PRINT("0x%x,", dataWord16[i]);
                  }
                }
              }
            }
          }
        }
      } else {
        LOGGER_PRINT("Ошибка ПДП(DMA) в ОУ(RT) (%s) = %d\n",
                     params.mNames[index], ret);
      }
      mil1553_close(fd);
    }
  }
}

void readDevices(void) {
  int index;
  // построение массива обнаруженных девайсов
  mil1553_getArrayDevices(&devices, &devicesCount);
  for (index = 0; index < devicesCount; index++) {
    devices[index].mDev.mFD =
        mil1553_openfl(devices[index].mDev.mNameChannel, O_RDWR);
    VERSION ver;
    mil1553_getDeviceInfo(devices[index].mDev.mFD, &ver);
    unsigned int verValue;
    mil1553_getDriverVersion(devices[index].mDev.mFD, &verValue);
    PCI_LOCATION loc;
    mil1553_getDevicePciInfo(devices[index].mDev.mFD, &loc);
    LOGGER_PRINT(
        "Девайс: '%s' deviceId= %d, revisionId= %d, pciloc= %s, verValue= "
        "0x%x\n",
        devices[index].mDev.mNameChannel, ver.device_id, ver.revision, loc.name,
        verValue);
  }
}

void readChannels(void) {
  int index;
  // построение массива обнаруженных каналов
  mil1553_getArrayChannels(&channels, &channelsCount);
  for (index = 0; index < channelsCount; index++) {
    int numDev = mil1553_getNumberDevice(&(channels[index]));
    LOGGER_PRINT("Девайс= %d - Канал: '%s'\n", numDev,
                 channels[index].mNameChannel);
  }
}

void printResults(void) {
  PRINT_LINE();
  LOGGER_PRINT_STR("Тест завершён.\n");
  if (TRUE_VAL == configChannels.mIsReady[0])
    LOGGER_PRINT("bc %s: dmablk= %d; err= %d;\n", params.mNames[0],
                 statTest.mCh[0].mDmaBlk, statTest.mCh[0].mErr);
  if (TRUE_VAL == configChannels.mIsReady[1])
    LOGGER_PRINT("rt1 %s: dmablk= %d; err= %d;\n", params.mNames[1],
                 statTest.mCh[1].mDmaBlk, statTest.mCh[1].mErr);
  if (TRUE_VAL == configChannels.mIsReady[2])
    LOGGER_PRINT("rt2 %s: dmablk= %d; err= %d;\n", params.mNames[2],
                 statTest.mCh[2].mDmaBlk, statTest.mCh[2].mErr);
  if (TRUE_VAL == configChannels.mIsReady[3])
    LOGGER_PRINT("rt3 %s: dmablk= %d; err= %d;\n", params.mNames[3],
                 statTest.mCh[3].mDmaBlk, statTest.mCh[3].mErr);
  if (TRUE_VAL == configChannels.mIsReady[4])
    LOGGER_PRINT("rt4 %s: dmablk= %d; err= %d;\n", params.mNames[4],
                 statTest.mCh[4].mDmaBlk, statTest.mCh[4].mErr);
  if (TRUE_VAL == configChannels.mIsReady[5])
    LOGGER_PRINT("rt5 %s: dmablk= %d; err= %d;\n", params.mNames[5],
                 statTest.mCh[5].mDmaBlk, statTest.mCh[5].mErr);
  if (TRUE_VAL == configChannels.mIsReady[6])
    LOGGER_PRINT("rt6 %s: dmablk= %d; err= %d;\n", params.mNames[6],
                 statTest.mCh[6].mDmaBlk, statTest.mCh[6].mErr);
  if (TRUE_VAL == configChannels.mIsReady[7])
    LOGGER_PRINT("rt7 %s: dmablk= %d; err= %d;\n", params.mNames[7],
                 statTest.mCh[7].mDmaBlk, statTest.mCh[7].mErr);
  if (TRUE_VAL == configChannels.mIsReady[8])
    LOGGER_PRINT("rt8 %s: dmablk= %d; err= %d;\n", params.mNames[8],
                 statTest.mCh[8].mDmaBlk, statTest.mCh[8].mErr);
}

int isValidNameChannels(void) {
  int ret = TRUE_VAL;
  int index, indexOfAllCh;
  for (index = 0; index < COUNT_CH; index++) {
    if (configChannels.mIsReady[index]) {
      int isGood = FALSE_VAL;
      for (indexOfAllCh = 0; indexOfAllCh < channelsCount; indexOfAllCh++) {
        if (0 == strncmp(params.mNames[index],
                         channels[indexOfAllCh].mNameChannel,
                         CHANNEL_NAME_LENGTH)) {
          isGood = TRUE_VAL;
          break;
        }
      }
      if (isGood == FALSE_VAL) {
        LOGGER_PRINT("Неправильное имя канала по индексу = %d, имя = '%s'\n",
                     index, params.mNames[index]);
        ret = FALSE_VAL;
      }
    }
  }
  return ret;
}

void bcDebugProcessing(void) {
  int index = 0;  // инрдекс КШ
  if (configChannels.mIsReady[index]) {
    SADDR_DATA regBcCondCode, regInstrPtr, regOpDataPtr;
    int fd = mil1553_open(params.mNames[index]);
    unsigned int numChannel = mil1553ud_getNumberChannel(fd);
    regBcCondCode.daddr = MIL_REG(numChannel, GPF_BC_COND_CODE_CH);
    regInstrPtr.daddr = MIL_REG(numChannel, START_ADR_INSTR_PTR_CH);
    regOpDataPtr.daddr = MIL_REG(numChannel, OP_DATA_RAM_PTR_CH);
    mil1553_readReg(fd, &regBcCondCode);
    mil1553_readReg(fd, &regInstrPtr);
    mil1553_readReg(fd, &regOpDataPtr);
    LOGGER_PRINT(
        "> REG_BC_COND_CODE: bcrun= %s, fifoerr= %s, operr= %s, insterr= %s, "
        "mbce= %s\n",
        ((regBcCondCode.data & (1u << 31)) ? "1" : "0"),
        ((regBcCondCode.data & (1u << 3)) ? "1" : "0"),
        ((regBcCondCode.data & (1u << 2)) ? "1" : "0"),
        ((regBcCondCode.data & (1u << 1)) ? "1" : "0"),
        ((regBcCondCode.data & (1u << 0)) ? "1" : "0"));
    SADDR_DATA instr;
    instr.daddr =
        MIL_REG(numChannel, BC_INSTR_RAM_CH) + ((regInstrPtr.data) & 0xFFF) * 4;
    mil1553_readReg(fd, &instr);
    LOGGER_PRINT(
        "  REG_INSTR_PTR: sinstr_ptr= 0x%x, cinstr_ptr= 0x%x, instruction= "
        "0x%x\n",
        (regInstrPtr.data >> 16) & 0xFFF, (regInstrPtr.data & 0xFFF),
        (instr.data, 16));
    SADDR_DATA oper;
    oper.daddr = MIL_REG(numChannel, BC_OPERATION_RAM_CH) +
                 ((regOpDataPtr.data >> 16) & 0xFFFF) * 4;
    mil1553_readReg(fd, &oper);
    LOGGER_PRINT(
        "  REG_OP_DATA_PTR: cop_ptr= 0x%x, cdat_ptr= 0x%x, operation= 0x%x\n",
        (regOpDataPtr.data >> 16) & 0xFFF, (regOpDataPtr.data & 0xFFF),
        (oper.data, 16));

    mil1553_close(fd);
  }
}

#define STEP_SECONDS(secstep, seccounter, tvPrev, tvCur, tvDiff) \
  gettimeofday(&tvCur, NULL);                                    \
  timersub(&tvCur, &tvPrev, &tvDiff);                            \
  if (tvDiff.tv_sec * 1000000 + tvDiff.tv_usec >=                \
      (secstep /*seconds*/ * 1000000)) {                         \
    tvPrev = tvCur;                                              \
    seccounter++;                                                \
    hasStep = MIL1553UD_TRUE;                                    \
  }

#define MAX_INTERRUPT_BLOCKS    64
INTERRUPT_BLOCK_BUFFER intBuffer;
struct block_info interruptBlocks[MAX_INTERRUPT_BLOCKS];

void testBcInt(void) {
  int countReadedBlocks = 0;
  int count = 0;
  int fd = 0;
  // LOGGER_PRINT("%s intBuffer ptr = 0x%x, intBuffer.blocks ptr = 0x%x \n", "КШ(BC) Чтение прерываний", &intBuffer, intBuffer.blocks);
  fd = mil1553_openfl(params.mNames[0], O_RDWR);
  do {
    count = mil1553ud_getCountInterruptBlocksReadyRead(fd);
    intBuffer.count_blocks = MAX_INTERRUPT_BLOCKS;
    unsigned int result = mil1553ud_readInterruptBlocks(fd, &intBuffer);
    if (result != GOOD) {
    LOGGER_PRINT("%s\n", "Ошибка функции mil1553ud_readInterruptBlocks");
    }
    countReadedBlocks += intBuffer.count_blocks;
  } while (count > 0);
  mil1553_close(fd);
  // LOGGER_PRINT("КШ(BC) - зафиксировано прерываний = %d; intBuffer ptr = 0x%x, intBuffer.blocks ptr = 0x%x %s", countReadedBlocks, &intBuffer, intBuffer.blocks, "\n");
  LOGGER_PRINT("КШ(BC) - зафиксировано прерываний = %d %s", countReadedBlocks, "\n");
}

void procCTest1(struct TParamCTest1 params) {
  int isValid;
  int countDevices;
  struct timeval tvPrev, tvCur, tvDiff;
  unsigned int allDurationSeconds;
  unsigned int curDurationSeconds;
  gettimeofday(&tvPrev, NULL);
  intBuffer.blocks = interruptBlocks;
  countDevices = mil1553_getCountDevices();
  if (countDevices == 0) {  // проверяем, что девайсов нет - значит и теста нет
    LOGGER_PRINT_STR(
        "Не найдено ни одного девайса. Невозможно провести тест.\n");
    goto END;
  }
  readDevices();   // печатаем список всех девайсов
  readChannels();  // печатаем список всех каналов
  isValid =
      isValidNameChannels();  // проверяем валидность имён каналов в параметрах
  if (isValid == FALSE_VAL) {  // если имена не валидны - теста нет
    LOGGER_PRINT_STR("Неверное имя канала.\n");
    goto END;
  }
  loadMemFiles();  // читаем и парсим мем файлы
  initTestDuration(
      &params, &allDurationSeconds,
      &curDurationSeconds);  // устанавливаем длительность теста из параметров
  stopBc();  // останавливаем КШ, если вдруг он с прошлого раза не остановлен
  deinitChannels();  // сброс всех каналов
  initChannels();  // ининциализация заданных каналов
  LOGGER_PRINT_STR("Подождите пока завершится тест, пожалуйста...\n");
  startBc();  // старт КШ
  isBcStopped = MIL1553UD_FALSE;
  int hasStep = MIL1553UD_TRUE;
  do {  // цикл теста
    if (MIL1553UD_TRUE ==
        hasStep) {  // раз в шаг печатаем отладочную информацию - это на случай
                    // если всё плохо
      // bcDebugProcessing();
      hasStep = MIL1553UD_FALSE;
      if (params.mBcInt) {
        testBcInt();
      }
    }
    //LOGGER_PRINT_STR("processingBc\n");
    processingBc();   // обработка блоков ДМА КШ
    //LOGGER_PRINT_STR("processingRts\n");
    processingRts();  // обработка блоков ДМА ОУ
    //LOGGER_PRINT_STR("STEP_SECONDS\n");
    STEP_SECONDS(1, curDurationSeconds, tvPrev, tvCur,
                 tvDiff);  // отсчитываем по одной секунде
    //LOGGER_PRINT_STR("end while\n");
  } while (curDurationSeconds <
           allDurationSeconds);  // проверка завершения теста по длительности
  LOGGER_PRINT_STR("stopBc\n");
  stopBc();                      // останов КШ
  isBcStopped = MIL1553UD_TRUE;
  unsigned int allDurationSecondsWait = 5;
  unsigned int curDurationSecondsWait = 0;
  if (configChannels.mIsReady[0]) {  // если КШ имеется
    gettimeofday(&tvPrev, NULL);
    do {  // ждём 5 секунд для остаточной отработки программы КШ
      processingBc();   // обработка блоков ДМА КШ
      processingRts();  // обработка блоков ДМА ОУ
      STEP_SECONDS(1, curDurationSecondsWait, tvPrev, tvCur, tvDiff);
    } while (curDurationSecondsWait < allDurationSecondsWait);
  }
  deinitChannels();  // сброс всех каналов
  printResults();    // печать результатов теста
  PRINT_LINE();
  LOGGER_PRINT("Длительность теста = %d секунд\n", allDurationSeconds);
END:
  return;
}
